# pyChaChaDummyProject (c) by chacha
#
# pyChaChaDummyProject is licensed under a
# Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <https://creativecommons.org/licenses/by-nc-sa/4.0/>.

from __future__ import annotations
from typing import TYPE_CHECKING

from pathlib import Path
import os
import datetime

import unittest
import xmlrunner
from junitparser import JUnitXml
from junit2htmlreport import parser as junit2html_parser

from .helper_base import helper_withresults_base


class unit_test(helper_withresults_base):
    enable_coverage_check: bool = False
    enable_xml_export: bool = True
    enable_full_xml_export: bool = True
    FullReportName: str = "full_report"
    CoverageReportName: str = "test_coverage"

    @classmethod
    def do_job(cls):
        if cls.enable_coverage_check == True:
            import coverage

        # preparing unittest framework
        test_loader = unittest.TestLoader()

        if cls.enable_coverage_check == True:
            # we start coverage now because module files discovery is part of the coverage measurement
            CoverageReportPath = Path(str(cls.get_result_dir()) + "_coverage")
            cls._reset_dir(CoverageReportPath)
            cov = coverage.Coverage(cover_pylib=False, branch=True, source_pkgs=["src." + cls.pyproject["project"]["name"]])
            cov.start()

        package_tests = test_loader.discover(
            start_dir=str(cls.project_rootdir_path / "test"), top_level_dir=str(cls.project_rootdir_path / "test")
        )
        if cls.enable_xml_export:
            testRunner = xmlrunner.XMLTestRunner(output=str(str(cls.get_result_dir())))
        else:
            testRunner = unittest.TextTestRunner()

        # running the test
        testRunner.run(package_tests)

        print("Test Finished")
        if cls.enable_coverage_check == True:
            cov.stop()
            cov.save()
            cov.html_report(directory=str(CoverageReportPath))
            cov.xml_report(outfile=(CoverageReportPath / f"{cls.CoverageReportName}.xml"))

        # computing results (Only if xml available)
        if cls.enable_full_xml_export == True:
            print("Full reports generation...")
            FullReportPath = Path(str(cls.get_result_dir()) + "_full")
            cls._reset_dir(FullReportPath)

            FullJUnitReport = JUnitXml()
            for fname in [fname for fname in os.listdir(cls.get_result_dir()) if fname.endswith(".xml")]:
                FullJUnitReport += JUnitXml.fromfile(str(cls.get_result_dir() / fname))

            current_datetime = datetime.datetime.utcnow().strftime("%Y%m%dT%H%M%SZ")
            full_report_base_name = f'{cls.pyproject["project"]["name"]}-{cls.FullReportName}-{current_datetime}'
            FullJUnitReport.write(str(FullReportPath / f"{full_report_base_name}.xml"))
            report = junit2html_parser.Junit(FullReportPath / f"{full_report_base_name}.xml")
            html = report.html()
            with open(FullReportPath / f"{full_report_base_name}.html", "wb") as outfile:
                outfile.write(html.encode("utf-8"))
            print("Done")



# pyChaChaDummyProject (c) by chacha
#
# pyChaChaDummyProject is licensed under a
# Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <https://creativecommons.org/licenses/by-nc-sa/4.0/>.

from __future__ import annotations
from typing import TYPE_CHECKING

from abc import ABC, abstractmethod
import os
from pathlib import Path
import subprocess

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from typing import Union


class helper_base(ABC):
    project_rootdir_path: Union[Path, None] = None
    pyproject: Union[dict, None] = None
    current_dir: Union[Path, None] = None

    @classmethod
    def set_context(cls, project_rootdir_path: Path, pyproject: dict):
        cls.project_rootdir_path = project_rootdir_path
        cls.pyproject = pyproject
        cls.current_dir = Path(__file__).parent.absolute()

    @classmethod
    def get_result_dir(cls):
        return None

    @staticmethod
    def _reset_dir(dirpath: Path):
        dirpath = Path(dirpath)
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        [f.unlink() for f in Path(dirpath).glob("*") if f.is_file()]

    @classmethod
    def reset_result_dir(cls):
        result_dir = cls.get_result_dir()
        if result_dir != None:
            cls._reset_dir(result_dir)

    @classmethod
    @abstractmethod
    def do_job(cls):
        raise NotImplementedError()

    @classmethod
    def run_cmd_(cls, cmdarray):
        process = subprocess.run(cmdarray, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, check=True)
        return process.stdout

    @classmethod
    def run_cmd(cls, cmdarray, silent: bool = False):
        p = subprocess.run(cmdarray, capture_output=True)
        if not silent:
            print(p.stdout.decode())
            print(p.stderr.decode())
        return p.stdout


class helper_withresults_base(helper_base):
    helper_results_dir: Union[Path, None] = None

    @classmethod
    def get_result_dir(cls):
        if cls.helper_results_dir == None:
            cls.helper_results_dir = cls.__name__
        return Path(__file__).parent.parent.absolute() / "helpers-results" / cls.helper_results_dir



# pyChaChaDummyProject (c) by chacha
#
# pyChaChaDummyProject is licensed under a
# Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <https://creativecommons.org/licenses/by-nc-sa/4.0/>.

from __future__ import annotations
from typing import TYPE_CHECKING

from pathlib import Path

from mypy import api

from .helper_base import helper_withresults_base


class types_check(helper_withresults_base):
    JUnitReportName = "junit.xml"

    @classmethod
    def do_job(cls):
        print("checking code typing ...")
        result = api.run(
            [  # project path
                "-p",
                "src." + cls.pyproject["project"]["name"],
                # analysis configuration
                # "--show-traceback",
                "--explicit-package-bases",
                # "--strict-equality",
                # "--check-untyped-defs",
                "--enable-incomplete-feature=Unpack",
                # reports generation
                "--cobertura-xml-report",
                str(cls.get_result_dir()),
                "--html-report",
                str(cls.get_result_dir()),
                "--txt-report",
                str(cls.get_result_dir()),
                "--xml-report",
                str(cls.get_result_dir()),
                "--junit-xml",
                str(cls.get_result_dir()) + "/" + cls.JUnitReportName,
            ]
        )

        if result[0]:
            print("\nType checking report:\n")
            print(result[0])  # stdout
            # converting the report using pylint_json2html (/!\ internal API, but as their is no leading '_' ...)
            with open(cls.get_result_dir() / "raw_eport.txt", "w+", encoding="utf-8") as Outfile:
                Outfile.write(result[0])

        if result[1]:
            print("\nError report:\n")
            print(result[1])  # stderr

        print("\nExit status:", result[2])
        print("Done")



# pygitversionhelper (c) by chacha
#
# pygitversionhelper  is licensed under a
# Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <https://creativecommons.org/licenses/by-nc-sa/4.0/>.

"""
Main module __init__ file.
"""

from importlib.metadata import version, PackageNotFoundError

try:  # pragma: no cover
    __version__ = version("pygitversionhelper")
except PackageNotFoundError:  # pragma: no cover
    import warnings

    warnings.warn("can not read __version__, assuming local test context, setting it to ?.?.?")
    __version__ = "?.?.?"

from .gitversionhelper import gitversionhelper, gitversionhelperException



# pygitversionhelper (c) by chacha
#
# pygitversionhelper  is licensed under a
# Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <https://creativecommons.org/licenses/by-nc-sa/4.0/>.

"""
This project try to help doing handy operations with git when
dealing with project versioning and tags on python project -
at leat for project using PEP440 or SemVer standards.

One requirement is to keep it compact and to not cover too much fancy features.
This is the reason why it is one single file with nested classes.

This library is made for repository that uses tags as version.
Support for non-version tags is optional and not well tested.

This module is the main project file, containing all the code.

Read the read me for more information.
Check the unittest s for usage samples.

///Note
_Other Parameters_ are **kwargs
///

Attributes:
    TKwargs: Kwargs type definition for type hints
"""

from __future__ import annotations

from typing import TYPE_CHECKING, cast, TypedDict, Literal
import os
import subprocess
import re
from copy import copy
import logging
from pathlib import Path

from packaging.version import VERSION_PATTERN as packaging_VERSION_PATTERN

if TYPE_CHECKING:
    from typing import List, Optional, Any, Dict
    from typing_extensions import Unpack

    TKwargs = TypedDict(
        "TKwargs",
        {
            "version_std": Optional[str],
            "same_branch": Optional[bool],
            "formated_output": Optional[bool],
            "bump_type": Optional[str],
            "bump_dev_strategy": Optional[str],
            "merged_output": Optional[bool],
            "ignore_unknown_tags": Optional[bool],
            "output_format": Optional[str],
            "ignore_merged": Optional[bool],
        },
        total=False,
    )


def _exec(cmd: str, root: Optional[str | os.PathLike[str]] = None, raw: bool = False) -> str | List[str]:
    """helper function to handle system cmd execution
    Args:
        cmd: command line to be executed
        root: root directory where the command need to be executed
        raw: return bytes if True, str if False or None
    Returns:
        a list of command's return lines or the raw output

    """
    _root: Path
    if isinstance(root, str):
        _root = Path(root)
    elif isinstance(root, os.PathLike):
        _root = Path(root)
    else:
        _root = Path(os.getcwd())

    p = subprocess.run(
        cmd,
        text=True,
        cwd=_root,
        capture_output=True,
        check=False,
        timeout=2,
        shell=True,
    )
    if re.search("not a git repository", p.stderr):
        raise gitversionhelper.repository.notAGitRepository()
    if re.search("fatal:", p.stderr):
        raise gitversionhelper.unknownGITFatalError(p.stderr)
    if int(p.returncode) < 0:
        raise gitversionhelper.unknownGITError(p.stderr)

    if raw is True:
        return p.stdout
    lines = p.stdout.splitlines()
    return [line.rstrip() for line in lines if line.rstrip()]


class gitversionhelperException(Exception):
    """general Module Exception"""


class gitversionhelper:  # pylint: disable=too-few-public-methods
    """main gitversionhelper class"""

    class wrongArguments(gitversionhelperException):
        """wrong argument generic exception"""

    class unknownGITError(gitversionhelperException):
        """unknown git error generic exception"""

    class unknownGITFatalError(unknownGITError):
        """unknown fatal git error generic exception"""

    class repository:
        """class containing methods focusing on repository"""

        class repositoryException(gitversionhelperException):
            """generic repository exeption"""

        class notAGitRepository(repositoryException):
            """not-a-git-repository repository exception"""

        class repositoryDirty(repositoryException):
            """dirty-repository repository exception"""

        @classmethod
        def isDirty(cls) -> bool:
            """check if the repository is in dirty state
            Returns:
                True if it is dirty
            """
            return bool(_exec("git status --short"))

    class commit:
        """class containing methods focusing on commits"""

        class commitException(gitversionhelperException):
            """generic commit exception"""

        class commitNotFound(commitException):
            """tag not found exception"""

        @classmethod
        def getMessagesSinceTag(cls, tag: str, **kwargs: Unpack[TKwargs]) -> str | List[str]:
            """Retrieve a commits message history from repository.

            Start from Latest found commit until the given tag.
            Args:
                tag: tag of the commit where search will stop
            Keyword Arguments:
                kwargs/merged_output (bool): Output one single merged string
                kwargs/same_branch (bool): Force searching only in the same branch
                kwargs/ignore_merged (bool): ignore merged commits
            Returns:
                the commit message
            """
            current_commit_id = cls.getLast(**kwargs)
            print(f"current_commit_id = {current_commit_id}")
            tag_commit_id = cls.getFromTag(tag)
            print(f"tag_commit_id = {tag_commit_id}")

            str_cmd: str
            if ("same_branch" in kwargs) and (kwargs["same_branch"] is True):
                str_cmd = f"git rev-list --first-parent {tag_commit_id}..{current_commit_id}"  # ok
            else:
                str_cmd = f"git rev-list {tag_commit_id}..{current_commit_id}"  # ok

            if ("ignore_merged" in kwargs) and (kwargs["ignore_merged"] is True):
                str_cmd = str_cmd + " --no-merges"  # ok

            try:
                commits = _exec(str_cmd)
            except gitversionhelper.unknownGITFatalError as _e:
                raise cls.commitNotFound("no commit found in commit history") from _e

            result = []
            for commit in commits:
                result.append(cls.getMessage(commit))

            if ("merged_output" in kwargs) and (kwargs["merged_output"] is True):
                return os.linesep.join(result)
            return result

        @classmethod
        def getMessage(cls, commit_hash: str) -> str:
            """retrieve a commit message from repository
            Args:
                commit_hash: id of the commit
            Returns:
                the commit message
            """
            try:
                res = _exec(
                    f'git log -z --pretty="tformat:%B%-C()" -n 1 {commit_hash}',  # ok
                    None,
                    True,
                )
                res = cast(str, res).rstrip("\x00")
            except gitversionhelper.unknownGITFatalError as _e:
                raise cls.commitNotFound("no commit found in commit history") from _e

            return res.replace("\r\n", "\n").replace("\n", "\r\n")

        @classmethod
        def getFromTag(cls, tag: str) -> str:
            """retrieve a commit from repository associated to a tag
            Args:
                tag: tag of the commit
            Returns:
                the commit Id
            """
            try:
                res = _exec(f"git rev-list -n 1 {tag}")  # ok
            except gitversionhelper.unknownGITFatalError as _e:
                raise cls.commitNotFound("no commit found in commit history") from _e

            if len(res) == 0:
                raise cls.commitNotFound("no commit found in commit history")
            return res[0]

        @classmethod
        def getLast(cls, **kwargs: Unpack[TKwargs]) -> str:
            """retrieve last commit from repository
            Keyword Arguments:
                kwargs/same_branch (bool): force searching only in the same branch
                kwargs/ignore_merged (bool): ignore merged commits
            Returns:
                the commit Id
            """
            str_cmd: str
            if ("same_branch" in kwargs) and (kwargs["same_branch"] is True):
                str_cmd = "git rev-list --max-count=1 --date-order HEAD --first-parent"  # ok
            else:
                str_cmd = "git log --format=%H --all -n1"  # ok

            if ("ignore_merged" in kwargs) and (kwargs["ignore_merged"] is True):
                str_cmd = str_cmd + " --no-merges"  # ok

            try:
                res = _exec(str_cmd)
            except gitversionhelper.unknownGITFatalError as _e:
                raise cls.commitNotFound("no commit found in commit history") from _e

            if len(res) == 0:
                raise cls.commitNotFound("no commit found in commit history")
            return res[0]

    class tag:
        """class containing methods focusing on tags"""

        __validGitTagSort = [
            "",
            "v:refname",
            "-v:refname",
            "taggerdate",
            "committerdate",
            "-taggerdate",
            "-committerdate",
        ]

        class tagException(gitversionhelperException):
            """generic tag exception"""

        class tagNotFound(tagException):
            """tag-not-found tag exception"""

        class moreThanOneTag(tagException):
            """more-than-one-tag tag exception"""

        @classmethod
        def getTags(cls, Sort: str = "taggerdate", **kwargs: Unpack[TKwargs]) -> List[str]:
            """retrieve all tags from a repository
            Args:
                Sort: sorting constraints (git format)
            Keyword Arguments:
                kwargs/same_branch (bool): force searching only in the same branch
            Returns:
                the tags list
            """

            if Sort not in cls.__validGitTagSort:
                raise gitversionhelper.wrongArguments("Sort option not in allowed list")

            if ("same_branch" in kwargs) and (kwargs["same_branch"] is True):
                currentBranch = _exec("git rev-parse --abbrev-ref HEAD")
                return list(reversed(_exec(f"git tag --merged {currentBranch[0]} --sort={Sort}")))
            return list(reversed(_exec(f"git tag -l --sort={Sort}")))

        @classmethod
        def getLastTag(cls, **kwargs: Unpack[TKwargs]) -> str:
            """retrieve the Latest tag from a repository
            Keyword Arguments:
                kwargs/same_branch (bool): force searching only in the same branch
            Returns:
                the tag
            """
            if ("same_branch" in kwargs) and (kwargs["same_branch"] is True):
                res = _exec("git describe --tags --first-parent --abbrev=0")
            else:
                res = _exec("git rev-list --tags --date-order --max-count=1")
                if len(res) == 1:
                    res = _exec(f"git describe --tags {res[0]}")

            if len(res) == 0:
                raise cls.tagNotFound("no tag found in commit history")
            if len(res) != 1:
                raise cls.moreThanOneTag("multiple tags on same commit is unsupported")
            return res[0]

        @classmethod
        def getDistanceFromTag(cls, tag: Optional[str] = None, **kwargs: Unpack[TKwargs]) -> int:
            """retrieve the distance between Latest commit and tag in the repository
            Arguments:
                tag: reference tag, if None the most recent one will be used
            Keyword Arguments:
                kwargs/same_branch (bool): force searching only in the same branch
            Returns:
                the tag
            """
            if tag is None:
                tag = cls.getLastTag(**kwargs)
            return int(_exec(f"git rev-list {tag}..HEAD --count")[0])

    class version:
        """class containing methods focusing on versions"""

        DefaultInputFormat = "Auto"

        TVersionStds = TypedDict("TVersionStds", {"regex": str, "regex_preversion_num": str, "regex_build_num": str}, total=False)

        VersionStds: dict[str, TVersionStds] = {
            "SemVer": {
                "regex": r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)"
                r"(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
                r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?"
                r"(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$",
                "regex_preversion_num": r"(?:\.)(?P<num>(?:\d+(?!\w))+)",
                "regex_build_num": r"(?:\.)(?P<num>(?:\d+(?!\w))+)",
            },
            "PEP440": {"regex": packaging_VERSION_PATTERN},
        }
        __versionReseted = False

        class versionException(gitversionhelperException):
            """generic version exception"""

        class noValidVersion(versionException):
            """no valid version found exception"""

        class PreAndPostVersionUnsupported(versionException):
            """pre and post release can not be present at the same time"""

        class MetaVersion:
            """generic version object"""

            TBumpTypes = Literal["major", "minor", "patch", "dev"]
            DefaultBumpType: TBumpTypes = "patch"
            BumpTypes: set[TBumpTypes] = {"major", "minor", "patch", "dev"}

            TBumpDevStrategies = Literal["post", "pre-patch", "pre-minor", "pre-major"]
            DefaultBumpDevStrategy: TBumpDevStrategies = "post"
            BumpDevStrategies: set[TBumpDevStrategies] = {"post", "pre-patch", "pre-minor", "pre-major"}

            TVersionStd = Literal["Auto", "PEP440", "SemVer"]
            version_std: TVersionStd = "Auto"
            major: int = 0
            minor: int = 1
            patch: int = 0
            pre_count: int = 0
            post_count: int = 0
            raw: str = "0.1.0"

            def __init__(
                self,
                version_std: TVersionStd = "Auto",
                major: int = 0,
                minor: int = 1,
                patch: int = 0,
                pre_count: int = 0,
                post_count: int = 0,
                raw: str = "0.1.0",
            ):  # pylint: disable=R0913
                self.version_std = version_std
                self.major = major
                self.minor = minor
                self.patch = patch
                self.pre_count = pre_count
                self.post_count = post_count
                self.raw = raw

            @classmethod
            def _getBumpDevStrategy(cls, **kwargs: Unpack[TKwargs]) -> str:
                """get selected bump_dev_strategy
                Keyword Arguments:
                    kwargs/bump_dev_strategy (str): the given bump_dev_strategy (can be None)
                Returns:
                    Kwargs given bump_dev_strategy or the default one.
                """
                BumpDevStrategy: str = cls.DefaultBumpDevStrategy

                if kwargs and ("bump_dev_strategy" in kwargs):
                    if kwargs["bump_dev_strategy"] in cls.BumpDevStrategies:
                        BumpDevStrategy = kwargs["bump_dev_strategy"]
                    else:
                        raise gitversionhelper.wrongArguments(f"invalid {'bump_type'} requested")
                return BumpDevStrategy

            @classmethod
            def _getBumpType(cls, **kwargs: Unpack[TKwargs]) -> str:
                """get selected bump_type
                Keyword Arguments:
                    kwargs/bump_type (str): the given bump_type (can be None)
                Returns:
                    Kwargs given bump_type or the default one.
                """
                BumpType: str = cls.DefaultBumpType
                if "bump_type" in kwargs:
                    if kwargs["bump_type"] in cls.BumpTypes:
                        BumpType = kwargs["bump_type"]
                    else:
                        raise gitversionhelper.wrongArguments(f"invalid {'bump_type'} requested")
                return BumpType

            def bump(  # pylint: disable=R1260,R0912
                self, amount: int = 1, **kwargs: Unpack[TKwargs]
            ) -> gitversionhelper.version.MetaVersion | str:
                """bump the version to the next one
                Args:
                    amount: number of revision to bump
                Keyword Arguments:
                    kwargs/bump_type (str): the given bump_type (can be None)
                    kwargs/bump_dev_strategy (str): the given bump_dev_strategy (can be None)
                Returns:
                    the bumped version
                """

                BumpType: str = self._getBumpType(**kwargs)

                BumpDevStrategy = self._getBumpDevStrategy(**kwargs)

                _v = copy(self)

                if BumpType == "dev":
                    if BumpDevStrategy == "post":
                        if _v.pre_count > 0:
                            _v.pre_count = _v.pre_count + amount
                        else:
                            _v.post_count = _v.post_count + amount
                    # elif BumpDevStrategy in ["pre-patch","pre-minor","pre-major"]:
                    else:
                        if _v.post_count > 0:
                            _v.post_count = _v.post_count + amount
                        else:
                            if _v.pre_count == 0:
                                if BumpDevStrategy == "pre-patch":
                                    _v.patch = _v.patch + 1
                                elif BumpDevStrategy == "pre-minor":
                                    _v.minor = _v.minor + 1
                                    _v.patch = 0
                                # elif BumpDevStrategy == "pre-major":
                                else:
                                    _v.major = _v.major + 1
                                    _v.minor = 0
                                    _v.patch = 0
                            _v.pre_count = _v.pre_count + amount
                else:
                    if BumpType == "major":
                        _v.major = _v.major + amount
                    elif BumpType == "minor":
                        _v.minor = _v.minor + amount
                    # elif BumpType == "patch":
                    else:
                        _v.patch = _v.patch + amount
                    _v.pre_count = 0
                    _v.post_count = 0

                _v.raw = _v.doFormatVersion(**kwargs)

                if ("formated_output" in kwargs) and (kwargs["formated_output"] is True):
                    return _v.doFormatVersion(**kwargs)
                return _v

            def doFormatVersion(self, **kwargs: Unpack[TKwargs]) -> str:
                """output a formated version string
                Keyword Arguments:
                    kwargs/output_format: output format to render ("Auto" or "PEP440" or "SemVer")
                Returns:
                    formated version string
                """
                return gitversionhelper.version.doFormatVersion(self, **kwargs)

        @classmethod
        def _getVersionStd(cls, **kwargs: Unpack[TKwargs]) -> gitversionhelper.version.MetaVersion.TVersionStd:
            """get selected version_std
            Keyword Arguments:
                kwargs/version_std (str): the given version_std (can be None)
            Returns:
                Kwargs given version_std or the default one.
            """
            VersionStd: str = cls.DefaultInputFormat
            if "version_std" in kwargs:
                if kwargs["version_std"] in cls.VersionStds:
                    VersionStd = kwargs["version_std"]
                else:
                    raise gitversionhelper.wrongArguments(f"invalid {'version_std'} requested")
            return cast(gitversionhelper.version.MetaVersion.TVersionStd, VersionStd)

        @classmethod
        def getCurrentVersion(cls, **kwargs: Unpack[TKwargs]) -> gitversionhelper.version.MetaVersion | str:
            """get the current version or bump depending of repository state.
            Keyword Arguments:
                kwargs/version_std (str): the given version_std (can be None)
                kwargs/same_branch (bool): force searching only in the same branch
                kwargs/formated_output (bool): output a formated version string
                kwargs/bump_type (str): the given bump_type (can be None)
                kwargs/bump_dev_strategy (str): the given bump_dev_strategy (can be None)
                kwargs/output_format (str): output format to render ("Auto" or "PEP440" or "SemVer")
            Returns:
                the last version
            """
            if gitversionhelper.repository.isDirty() is not False:
                raise gitversionhelper.repository.repositoryDirty("The repository is dirty and a current version can not be generated.")

            saved_kwargs = copy(kwargs)
            if "formated_output" in kwargs:
                del saved_kwargs["formated_output"]

            _v = cast(gitversionhelper.version.MetaVersion, cls.getLastVersion(**saved_kwargs))

            if not cls.__versionReseted:
                amount = gitversionhelper.tag.getDistanceFromTag(_v.raw, **saved_kwargs)
                _v = cast(gitversionhelper.version.MetaVersion, _v.bump(amount, **saved_kwargs))

            if ("formated_output" in kwargs) and (kwargs["formated_output"] is True):
                return _v.doFormatVersion(**kwargs)
            return _v

        @classmethod
        def getCurrentFormatedVersion(cls, **kwargs: Unpack[TKwargs]) -> str:
            """same as getCurrentVersion() with formated_output kwarg forced activated.
            Keyword Arguments:
                kwargs/version_std (str): the given version_std (can be None)
                kwargs/same_branch (bool): force searching only in the same branch
                kwargs/bump_type (str): the given bump_type (can be None)
                kwargs/bump_dev_strategy (str): the given bump_dev_strategy (can be None)
                kwargs/output_format (str): output format to render ("Auto" or "PEP440" or "SemVer")
            Returns:
                the last version
            """
            kwargs["formated_output"] = True
            return cast(str, cls.getCurrentVersion(**kwargs))

        @classmethod
        def _parseTag(  # pylint: disable=R1260,R0914,R0912,R0915
            cls, tag: str, **kwargs: Unpack[TKwargs]
        ) -> gitversionhelper.version.MetaVersion:
            """get version from tags.
            Arguments:
                tag: the tag to be parsed
            Keyword Arguments:
                kwargs/version_std (str): the given version_std (can be None)
            Returns:
                MetaVersion object
            """
            _m: Optional[re.Match[str]]
            VersionStd: gitversionhelper.version.MetaVersion.TVersionStd = cls._getVersionStd(**kwargs)
            bAutoVersionStd = False
            if VersionStd == "Auto":
                bAutoVersionStd = True
            bFound = False
            if VersionStd == "SemVer" or (bAutoVersionStd is True):
                _r = re.compile(
                    r"^\s*" + cls.VersionStds["SemVer"]["regex"] + r"\s*$",
                    re.VERBOSE | re.IGNORECASE,
                )
                _m = re.match(_r, tag)
                if _m is None:
                    pass
                else:
                    major = int(_m.group("major"))
                    minor = int(_m.group("minor"))
                    patch = int(_m.group("patch"))

                    pre_count = 0
                    if _pre := _m.group("prerelease"):
                        if (_match := re.search(cls.VersionStds["SemVer"]["regex_preversion_num"], _pre)) is not None:
                            pre_count = int(_match.group("num"))
                        else:
                            pre_count = 1

                    post_count = 0
                    if _post := _m.group("buildmetadata"):
                        if (_match := re.search(cls.VersionStds["SemVer"]["regex_build_num"], _post)) is not None:
                            post_count = int(_match.group("num"))
                        else:
                            post_count = 1
                    bFound = True
                    VersionStd = "SemVer"

            if VersionStd == "PEP440" or ((bAutoVersionStd is True) and (bFound is not True)):
                _r = re.compile(
                    r"^\s*" + cls.VersionStds["PEP440"]["regex"] + r"\s*$",
                    re.VERBOSE | re.IGNORECASE,
                )
                _m = re.match(_r, tag)
                if _m is None:
                    pass
                else:
                    res: str = _m.group("release")
                    if isinstance(res, str):

                        ver = res.split(".")
                        ver += ["0"] * (3 - len(ver))

                        ver_int: List[int] = [0, 0, 0]
                        ver_int[0] = int(ver[0])
                        ver_int[1] = int(ver[1])
                        ver_int[2] = int(ver[2])
                        major, minor, patch = tuple(ver_int)

                        pre_count = int(_m.group("pre_n")) if _m.group("pre_n") else 0
                        post_count = int(_m.group("post_n2")) if _m.group("post_n2") else 0
                        bFound = True
                        VersionStd = "PEP440"

            if not bFound:
                raise gitversionhelper.version.noValidVersion("no valid version found in tags")

            if pre_count > 0 and post_count > 0:
                raise cls.PreAndPostVersionUnsupported("can not parse a version with both pre and post release number.")
            return cls.MetaVersion(VersionStd, major, minor, patch, pre_count, post_count, tag)

        @classmethod
        def getLastVersion(cls, **kwargs: Unpack[TKwargs]) -> gitversionhelper.version.MetaVersion | str:  # pylint: disable=R1260
            """get the last version from tags
            Keyword Arguments:
                kwargs/version_std (str): the given version_std (can be None)
                kwargs/same_branch (bool): force searching only in the same branch
                kwargs/formated_output (bool): output a formated version string
                kwargs/ignore_unknown_tags (bool): skip tags with not decoded versions (default to False)
            Returns:
                the last version in MetaVersion object or string
            """
            lastTag: str = cls.MetaVersion.raw
            cls.__versionReseted = False
            try:
                lastTag = cast(str, gitversionhelper.tag.getLastTag(**kwargs))
            except gitversionhelper.tag.tagNotFound:
                logging.warning("tag not found, reseting versionning")
                cls.__versionReseted = True

            _v: Optional[gitversionhelper.version.MetaVersion] = None
            try:
                _v = cls._parseTag(lastTag, **kwargs)
            except gitversionhelper.version.noValidVersion as _ex:
                if ("ignore_unknown_tags" in kwargs) and (kwargs["ignore_unknown_tags"] is True):
                    tags = gitversionhelper.tag.getTags(Sort="taggerdate", **kwargs)
                    _v = None
                    for _tag in tags:
                        try:
                            _v = cls._parseTag(_tag, **kwargs)
                            break
                        except gitversionhelper.version.noValidVersion:
                            continue
                if _v is None:
                    raise gitversionhelper.version.noValidVersion() from _ex

            if ("formated_output" in kwargs) and (kwargs["formated_output"] is True):
                return _v.doFormatVersion(**kwargs)
            return _v

        @classmethod
        def doFormatVersion(cls, inputversion: MetaVersion, **kwargs: Unpack[TKwargs]) -> str:
            """output a formated version string from a MetaVersion object
            Keyword Arguments:
                kwargs/output_format (str): output format to render ("Auto" or "PEP440" or "SemVer")
            Args:
                inputversion: version to be rendered
            Returns:
                formated version string
            """

            VersionStd = cls._getVersionStd(**kwargs)
            if VersionStd == "Auto":
                VersionStd = inputversion.version_std

            revpattern = ""
            revcount = ""
            post_count = inputversion.post_count
            pre_count = inputversion.pre_count
            patch = inputversion.patch

            OutputFormat = kwargs.get("output_format")

            if OutputFormat is None:
                OutputFormat = "{major}.{minor}.{patch}{revpattern}{revcount}"
                if post_count > 0 and pre_count > 0:
                    raise gitversionhelper.version.PreAndPostVersionUnsupported(
                        "cannot output a version with both pre and post release number."
                    )
                if VersionStd == "PEP440":
                    if post_count > 0:
                        revpattern = ".post"
                        revcount = f"{post_count}"
                    elif pre_count > 0:
                        revpattern = ".pre"
                        revcount = f"{pre_count}"
                # elif    VersionStd == "SemVer":
                else:
                    if post_count > 0:
                        revpattern = "+post"
                        revcount = f".{post_count}"
                    elif pre_count > 0:
                        revpattern = "-pre"
                        revcount = f".{pre_count}"
            return OutputFormat.format(
                major=inputversion.major,
                minor=inputversion.minor,
                patch=patch,
                revpattern=revpattern,
                revcount=revcount,
            )



# pygitversionhelper (c) by chacha
#
# pygitversionhelper  is licensed under a
# Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <https://creativecommons.org/licenses/by-nc-sa/4.0/>.

import unittest

import tempfile
import os
import pathlib
import re
import copy
import time
import subprocess

from src import pygitversionhelper

HelperRegex = r"^(?P<MAJ>\d+)\.(?P<MIN>\d+)\.(?P<PATCH>\d+)([\.\-\+])?(?:.*)?"


class Test_gitversionhelper(unittest.TestCase):
    def setUp(self):
        self.TmpWorkingDir = tempfile.TemporaryDirectory()
        self.TmpWorkingDirPath = pathlib.Path(self.TmpWorkingDir.name)
        os.chdir(self.TmpWorkingDirPath)
        os.system("git init")
        os.system('git config --local user.name "john doe"')
        os.system('git config --local user.email "john@doe.org"')

    def _test_version_readback(self, tag: str, **kwargs):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")

        os.system("git add .")
        os.system('git commit -m  "first commit"')
        os.system(f"git tag {tag}")

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(**kwargs)

        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self.assertIsInstance(_v.raw, str)
        self.assertEqual(_v.raw, tag)
        self.assertIsInstance(_v.major, int)
        self.assertIsInstance(_v.minor, int)
        self.assertIsInstance(_v.patch, int)
        self.assertIsInstance(_v.pre_count, int)
        self.assertIsInstance(_v.post_count, int)

        _m = re.match(HelperRegex, tag)
        self.assertEqual(int(_m.group("MAJ")), _v.major)
        self.assertEqual(int(_m.group("MIN")), _v.minor)
        self.assertEqual(int(_m.group("PATCH")), _v.patch)

        return _v

    def _test_version_format(
        self,
        _v: pygitversionhelper.gitversionhelper.version.MetaVersion,
        tag: str,
        **kwargs,
    ):
        _f = _v.doFormatVersion(**kwargs)
        self.assertRegex(_f, HelperRegex)
        self.assertEqual(_f, tag)

    def _test_version_readback_simple(self, tag: str, **kwargs):
        _v = self._test_version_readback(tag, **kwargs)
        self._test_version_format(_v, tag, **kwargs)

    def test_nominal__version__formated_output(self):
        _v = pygitversionhelper.gitversionhelper.version.MetaVersion("PEP440", 1, 0, 0, 0, 0, "1.0.0")
        self.assertEqual("1.0.1", _v.bump(formated_output=True))
        self.assertEqual("2.0.0", _v.bump(formated_output=True, bump_type="major"))

    def test_nominal__version__auto_1(self):
        self._test_version_readback_simple("0.0.1")

    def test_nominal__version__auto_2(self):
        self._test_version_readback_simple("0.0.2")

    def test_nominal__version__auto_3(self):
        self._test_version_readback_simple("0.1.0")

    def test_nominal__version__auto_4(self):
        self._test_version_readback_simple("0.1.1")

    def test_nominal__version__auto_5(self):
        self._test_version_readback_simple("1.0.0")

    def test_nominal__version__auto_6(self):
        self._test_version_readback_simple("1.1.0")

    def test_nominal__version__auto_7(self):
        self._test_version_readback_simple("1.2.0")

    def test_nominal__version__auto_8(self):
        self._test_version_readback_simple("1.1.1")

    def test_nominal__version__auto_9(self):
        self._test_version_readback_simple("1.2.1")

    def test_nominal__version__auto_PEP440_post(self):
        self._test_version_readback_simple("1.2.1.post1")

    def test_nominal__version__auto_PEP440_pre(self):
        self._test_version_readback_simple("1.2.1.pre1")

    def test_nominal__version__auto_PEP440_post_2(self):
        self._test_version_readback_simple("1.2.1.post10")

    def test_nominal__version__auto_PEP440_pre_2(self):
        self._test_version_readback_simple("1.2.1.pre10")

    def test_nominal__version__auto_SemVer_post(self):
        self._test_version_readback_simple("1.2.1+post.1")

    def test_nominal__version__auto_SemVer_pre(self):
        self._test_version_readback_simple("1.2.1-pre.1")

    def test_nominal__version__auto_SemVer_post_2(self):
        self._test_version_readback_simple("1.2.1+post.10")

    def test_nominal__version__auto_Semver_pre_2(self):
        self._test_version_readback_simple("1.2.1-pre.10")

    def test_nominal__version__PEP440_1(self):
        self._test_version_readback_simple("0.0.1", version_std="PEP440")

    def test_nominal__version__PEP440_2(self):
        self._test_version_readback_simple("0.0.2", version_std="PEP440")

    def test_nominal__version__PEP440_3(self):
        self._test_version_readback_simple("0.1.0", version_std="PEP440")

    def test_nominal__version__PEP440_4(self):
        self._test_version_readback_simple("0.1.1", version_std="PEP440")

    def test_nominal__version__PEP440_5(self):
        self._test_version_readback_simple("1.0.0", version_std="PEP440")

    def test_nominal__version__PEP440_6(self):
        self._test_version_readback_simple("1.1.0", version_std="PEP440")

    def test_nominal__version__PEP440_7(self):
        self._test_version_readback_simple("1.2.0", version_std="PEP440")

    def test_nominal__version__PEP440_8(self):
        self._test_version_readback_simple("1.1.1", version_std="PEP440")

    def test_nominal__version__PEP440_9(self):
        self._test_version_readback_simple("1.2.1", version_std="PEP440")

    def test_nominal__version__PEP440_post(self):
        self._test_version_readback_simple("1.2.1.post1", version_std="PEP440")

    def test_nominal__version__PEP440_pre(self):
        self._test_version_readback_simple("1.2.1.pre1", version_std="PEP440")

    def test_nominal__version__PEP440_post_2(self):
        self._test_version_readback_simple("1.2.1.post10", version_std="PEP440")

    def test_nominal__version__PEP440_pre_2(self):
        self._test_version_readback_simple("1.2.1.pre10", version_std="PEP440")

    def test_nominal__version__SemVer_1(self):
        self._test_version_readback_simple("0.0.1", version_std="SemVer")

    def test_nominal__version__SemVer_2(self):
        self._test_version_readback_simple("0.0.2", version_std="SemVer")

    def test_nominal__version__SemVer_3(self):
        self._test_version_readback_simple("0.1.0", version_std="SemVer")

    def test_nominal__version__SemVer_4(self):
        self._test_version_readback_simple("0.1.1", version_std="SemVer")

    def test_nominal__version__SemVer_5(self):
        self._test_version_readback_simple("1.0.0", version_std="SemVer")

    def test_nominal__version__SemVer_6(self):
        self._test_version_readback_simple("1.1.0", version_std="SemVer")

    def test_nominal__version__SemVer_7(self):
        self._test_version_readback_simple("1.2.0", version_std="SemVer")

    def test_nominal__version__SemVer_8(self):
        self._test_version_readback_simple("1.1.1", version_std="SemVer")

    def test_nominal__version__SemVer_9(self):
        self._test_version_readback_simple("1.2.1", version_std="SemVer")

    def test_nominal__version__SemVer_post(self):
        self._test_version_readback_simple("1.2.1+post.1", version_std="SemVer")

    def test_nominal__version__SemVer_pre(self):
        self._test_version_readback_simple("1.2.1-pre.1", version_std="SemVer")

    def test_nominal__version__SemVer_post_2(self):
        self._test_version_readback_simple("1.2.1+post.10", version_std="SemVer")

    def test_nominal__version__SemVer_pre_2(self):
        self._test_version_readback_simple("1.2.1-pre.10", version_std="SemVer")

    def test_nominal__version__post_PEP440_to_SemVer(self):
        _v = self._test_version_readback("1.2.1.post10", version_std="PEP440")
        self._test_version_format(_v, "1.2.1+post.10", version_std="SemVer")

    def test_nominal__version__post_SemVer_to_PEP440(self):
        _v = self._test_version_readback("1.2.1+post.10", version_std="SemVer")
        self._test_version_format(_v, "1.2.1.post10", version_std="PEP440")

    def test_nominal__version__pre_PEP440_to_SemVer(self):
        _v = self._test_version_readback("1.2.1.pre10", version_std="PEP440")
        self._test_version_format(_v, "1.2.1-pre.10", version_std="SemVer")

    def test_nominal__version__pre_SemVer_to_PEP440(self):
        _v = self._test_version_readback("1.2.1-pre.10", version_std="SemVer")
        self._test_version_format(_v, "1.2.1.pre10", version_std="PEP440")

    def test_nominal__version__post_SemVer_nonum(self):
        _v = self._test_version_readback("1.2.1+post", version_std="SemVer")
        self._test_version_format(_v, "1.2.1+post.1", version_std="SemVer")

    def test_nominal__version__post_SemVer_nonum_random(self):
        _v = self._test_version_readback("1.2.1+toto", version_std="SemVer")
        self._test_version_format(_v, "1.2.1+post.1", version_std="SemVer")

    def test_nominal__version__pre_SemVer_nonum(self):
        _v = self._test_version_readback("1.2.1-pre", version_std="SemVer")
        self._test_version_format(_v, "1.2.1-pre.1", version_std="SemVer")

    def test_nominal__version__pre_SemVer_nonum_random(self):
        _v = self._test_version_readback("1.2.1-toto", version_std="SemVer")
        self._test_version_format(_v, "1.2.1-pre.1", version_std="SemVer")

    def test_nominal__version___pump_SemVer(self):
        _v = self._test_version_readback("1.0.0", version_std="SemVer")

        _v = _v.bump()
        self.assertIsInstance(_v.raw, str)
        self.assertIsInstance(_v.major, int)
        self.assertIsInstance(_v.minor, int)
        self.assertIsInstance(_v.patch, int)
        self.assertIsInstance(_v.pre_count, int)
        self.assertIsInstance(_v.post_count, int)
        self.assertEqual(_v.raw, "1.0.1")
        self.assertEqual(_v.major, 1)
        self.assertEqual(_v.minor, 0)
        self.assertEqual(_v.patch, 1)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "1.0.1")

        _v = _v.bump(bump_type="patch")
        self.assertEqual(_v.raw, "1.0.2")
        self.assertEqual(_v.major, 1)
        self.assertEqual(_v.minor, 0)
        self.assertEqual(_v.patch, 2)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "1.0.2")

        _v = _v.bump(bump_type="minor")
        self.assertEqual(_v.raw, "1.1.2")
        self.assertEqual(_v.major, 1)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 2)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "1.1.2")

        _v = _v.bump(bump_type="major")
        self.assertEqual(_v.raw, "2.1.2")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 2)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.2")

        _v = _v.bump(bump_type="dev")
        self.assertEqual(_v.raw, "2.1.2+post.1")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 2)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 1)
        self.assertEqual(_v.doFormatVersion(), "2.1.2+post.1")

        _v = _v.bump(bump_type="dev")
        self.assertEqual(_v.raw, "2.1.2+post.2")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 2)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 2)
        self.assertEqual(_v.doFormatVersion(), "2.1.2+post.2")
        self.assertEqual(_v.doFormatVersion(version_std="SemVer"), "2.1.2+post.2")
        self.assertEqual(_v.doFormatVersion(version_std="PEP440"), "2.1.2.post2")

        _v = _v.bump(bump_type="patch")
        self.assertEqual(_v.raw, "2.1.3")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 3)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.3")

        _v = _v.bump(bump_type="dev", bump_dev_strategy="pre-patch")
        self.assertEqual(_v.raw, "2.1.4-pre.1")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 4)
        self.assertEqual(_v.pre_count, 1)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.4-pre.1")

        _v = _v.bump(bump_type="dev", bump_dev_strategy="pre-patch")
        self.assertEqual(_v.raw, "2.1.4-pre.2")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 4)
        self.assertEqual(_v.pre_count, 2)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.4-pre.2")

        _v = _v.bump(bump_type="dev", bump_dev_strategy="post")
        self.assertEqual(_v.raw, "2.1.4-pre.3")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 4)
        self.assertEqual(_v.pre_count, 3)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.4-pre.3")

        _v = _v.bump(bump_type="patch")
        self.assertEqual(_v.raw, "2.1.5")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 5)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.5")

        _v = _v.bump(bump_type="dev", bump_dev_strategy="post")
        self.assertEqual(_v.raw, "2.1.5+post.1")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 5)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 1)
        self.assertEqual(_v.doFormatVersion(), "2.1.5+post.1")

        _v = _v.bump(bump_type="dev", bump_dev_strategy="pre-patch")
        self.assertEqual(_v.raw, "2.1.5+post.2")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 5)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 2)
        self.assertEqual(_v.doFormatVersion(), "2.1.5+post.2")

    def test_nominal__version___pump_PEP440(self):
        _v = self._test_version_readback("1.0.0", version_std="PEP440")

        _v = _v.bump()
        self.assertIsInstance(_v.raw, str)
        self.assertIsInstance(_v.major, int)
        self.assertIsInstance(_v.minor, int)
        self.assertIsInstance(_v.patch, int)
        self.assertIsInstance(_v.pre_count, int)
        self.assertIsInstance(_v.post_count, int)
        self.assertEqual(_v.raw, "1.0.1")
        self.assertEqual(_v.major, 1)
        self.assertEqual(_v.minor, 0)
        self.assertEqual(_v.patch, 1)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "1.0.1")

        _v = _v.bump(bump_type="patch")
        self.assertEqual(_v.raw, "1.0.2")
        self.assertEqual(_v.major, 1)
        self.assertEqual(_v.minor, 0)
        self.assertEqual(_v.patch, 2)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "1.0.2")

        _v = _v.bump(bump_type="minor")
        self.assertEqual(_v.raw, "1.1.2")
        self.assertEqual(_v.major, 1)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 2)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "1.1.2")

        _v = _v.bump(bump_type="major")
        self.assertEqual(_v.raw, "2.1.2")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 2)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.2")

        _v = _v.bump(bump_type="dev")
        self.assertEqual(_v.raw, "2.1.2.post1")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 2)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 1)
        self.assertEqual(_v.doFormatVersion(), "2.1.2.post1")

        _v = _v.bump(bump_type="dev")
        self.assertEqual(_v.raw, "2.1.2.post2")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 2)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 2)
        self.assertEqual(_v.doFormatVersion(), "2.1.2.post2")

        _v = _v.bump(bump_type="patch")
        self.assertEqual(_v.raw, "2.1.3")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 3)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.3")

        _v = _v.bump(bump_type="dev", bump_dev_strategy="pre-patch")
        self.assertEqual(_v.raw, "2.1.4.pre1")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 4)
        self.assertEqual(_v.pre_count, 1)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.4.pre1")

        _v = _v.bump(bump_type="dev", bump_dev_strategy="pre-patch")
        self.assertEqual(_v.raw, "2.1.4.pre2")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 4)
        self.assertEqual(_v.pre_count, 2)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.4.pre2")

        _v = _v.bump(bump_type="dev", bump_dev_strategy="post")
        self.assertEqual(_v.raw, "2.1.4.pre3")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 4)
        self.assertEqual(_v.pre_count, 3)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.4.pre3")

        _v = _v.bump(bump_type="patch")
        self.assertEqual(_v.raw, "2.1.5")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 5)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)
        self.assertEqual(_v.doFormatVersion(), "2.1.5")

        _v = _v.bump(bump_type="dev", bump_dev_strategy="post")
        self.assertEqual(_v.raw, "2.1.5.post1")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 5)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 1)
        self.assertEqual(_v.doFormatVersion(), "2.1.5.post1")

        _v = _v.bump(bump_type="dev", bump_dev_strategy="pre-patch")
        self.assertEqual(_v.raw, "2.1.5.post2")
        self.assertEqual(_v.major, 2)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 5)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 2)
        self.assertEqual(_v.doFormatVersion(), "2.1.5.post2")

    def test_nominal__version___SemVer_zeroRev(self):
        _v = self._test_version_readback("0.0.0", version_std="SemVer")
        self._test_version_format(_v, "0.0.0", version_std="SemVer")

    def test_nominal__version___PEP440_zeroRev(self):
        _v = self._test_version_readback("0.0.0", version_std="PEP440")
        self._test_version_format(_v, "0.0.0", version_std="PEP440")

    def test_nominal__version___PEP440_noRev_noTag(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")

        os.system("git add .")
        os.system('git commit -m  "first commit"')

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(version_std="PEP440")

        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.1.0", version_std="PEP440")
        self.assertEqual(_v.raw, "0.1.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 0)

    def test_nominal__version___SemVer_noRev_noTag(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")

        os.system("git add .")
        os.system('git commit -m  "first commit"')

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(version_std="SemVer")

        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.1.0", version_std="SemVer")
        self.assertEqual(_v.raw, "0.1.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 0)

    def test_nominal__version___AUTO_noRev_noTag(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")

        os.system("git add .")
        os.system('git commit -m  "first commit"')

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion()

        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.1.0")
        self.assertEqual(_v.raw, "0.1.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 0)

    def test_nominal__version___getCurrentFormatedVersion(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system('git commit -m  "first commit"')
        os.system(f"git tag 0.2.0")

        self.assertEqual(pygitversionhelper.gitversionhelper.version.getCurrentFormatedVersion(), "0.2.0")

    def test_nominal__version___AUTO_bump_commits(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system('git commit -m  "first commit"')
        os.system(f"git tag 0.2.0")

        _v = pygitversionhelper.gitversionhelper.version.getCurrentVersion()

        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.2.0")
        self.assertEqual(_v.raw, "0.2.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 2)
        self.assertEqual(_v.patch, 0)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue22")
        os.system("git add .")
        os.system('git commit -m  "2nd commit"')

        _v = pygitversionhelper.gitversionhelper.version.getCurrentVersion()

        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.2.1")
        self.assertEqual(_v.raw, "0.2.1")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 2)
        self.assertEqual(_v.patch, 1)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue223")
        os.system("git add .")
        os.system('git commit -m  "3rd commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue224")
        os.system("git add .")
        os.system('git commit -m  "4th commit"')

        _v = pygitversionhelper.gitversionhelper.version.getCurrentVersion()

        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.2.3")
        self.assertEqual(_v.raw, "0.2.3")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 2)
        self.assertEqual(_v.patch, 3)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue225")
        os.system("git add .")
        os.system('git commit -m  "5th commit"')

        _v = pygitversionhelper.gitversionhelper.version.getCurrentVersion(bump_type="dev")

        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self.assertIsInstance(_v.raw, str)
        self._test_version_format(_v, "0.2.0+post.4")
        self.assertEqual(_v.raw, "0.2.0+post.4")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 2)
        self.assertEqual(_v.patch, 0)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 4)

        _v = pygitversionhelper.gitversionhelper.version.getCurrentVersion(bump_type="dev", bump_dev_strategy="pre-patch")

        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.2.1-pre.4")
        self.assertEqual(_v.raw, "0.2.1-pre.4")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 2)
        self.assertEqual(_v.patch, 1)
        self.assertEqual(_v.pre_count, 4)
        self.assertEqual(_v.post_count, 0)

        _v = pygitversionhelper.gitversionhelper.version.getCurrentVersion(bump_type="dev", bump_dev_strategy="pre-minor")

        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.3.0-pre.4")
        self.assertEqual(_v.raw, "0.3.0-pre.4")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 3)
        self.assertEqual(_v.patch, 0)
        self.assertEqual(_v.pre_count, 4)
        self.assertEqual(_v.post_count, 0)

        _v = pygitversionhelper.gitversionhelper.version.getCurrentVersion(bump_type="dev", bump_dev_strategy="pre-major")

        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "1.0.0-pre.4")
        self.assertEqual(_v.raw, "1.0.0-pre.4")
        self.assertEqual(_v.major, 1)
        self.assertEqual(_v.minor, 0)
        self.assertEqual(_v.patch, 0)
        self.assertEqual(_v.pre_count, 4)
        self.assertEqual(_v.post_count, 0)

    def test_nominal__version___custom_format(self):
        _v = self._test_version_readback("0.1.1", version_std="PEP440")

        self.assertEqual(
            "V_0_1_1____",
            _v.doFormatVersion(output_format="V_{major}_{minor}_{patch}____{revcount}"),
        )
        self.assertEqual("V_0_1", _v.doFormatVersion(output_format="V_{major}_{minor}"))

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue225")
        os.system("git add .")
        os.system('git commit -m  "5th commit"')

        self.assertEqual(
            "V_1_0",
            pygitversionhelper.gitversionhelper.version.getCurrentVersion(bump_type="dev", bump_dev_strategy="pre-major").doFormatVersion(
                output_format="V_{major}_{minor}"
            ),
        )
        self.assertEqual(
            "V_1_0",
            pygitversionhelper.gitversionhelper.version.getCurrentVersion(
                bump_type="dev",
                bump_dev_strategy="pre-major",
                output_format="V_{major}_{minor}",
                formated_output=True,
            ),
        )
        self.assertEqual(
            "0.1.1",
            pygitversionhelper.gitversionhelper.version.getLastVersion(
                formated_output=True,
            ),
        )

    def test_nominal__git__emptyrepo(self):
        _v = pygitversionhelper.gitversionhelper.version.getCurrentVersion()
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.1.0")
        self.assertEqual(_v.raw, "0.1.0")

        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 0)
        self.assertEqual(_v.pre_count, 0)
        self.assertEqual(_v.post_count, 0)

    def test_defect__git__dirty(self):
        _v = self._test_version_readback("0.1.1", version_std="PEP440")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue225")

        with self.assertRaises(pygitversionhelper.gitversionhelper.repository.repositoryDirty):
            pygitversionhelper.gitversionhelper.version.getCurrentVersion()

    def test_defect__version_post_and_pre_output(self):
        _v = pygitversionhelper.gitversionhelper.version.MetaVersion("PEP440", 0, 0, 1, 1, 1, "0.0.1.pre1.post1")
        with self.assertRaises(pygitversionhelper.gitversionhelper.version.PreAndPostVersionUnsupported):
            _v.doFormatVersion()

        _v = pygitversionhelper.gitversionhelper.version.MetaVersion("SemVer", 0, 0, 1, 1, 1, "0.0.1-pre.1+post.1")
        with self.assertRaises(pygitversionhelper.gitversionhelper.version.PreAndPostVersionUnsupported):
            _v.doFormatVersion()

    def test_defect__version_post_and_pre_parse(self):
        with self.assertRaises(pygitversionhelper.gitversionhelper.version.PreAndPostVersionUnsupported):
            pygitversionhelper.gitversionhelper.version._parseTag("0.0.1.pre1.post1")

    def test_defect__git__wrongargument_Sortargs(self):
        with self.assertRaises(pygitversionhelper.gitversionhelper.wrongArguments):
            pygitversionhelper.gitversionhelper.tag.getTags(Sort="toto")

    def test_defect__git__notagfound(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system('git commit -m  "first commit"')
        with self.assertRaises(pygitversionhelper.gitversionhelper.tag.tagNotFound):
            pygitversionhelper.gitversionhelper.tag.getLastTag()

    """ This test is impossible to do because current implementation can only return one tag
    def test_defect__git_multipletagsfound(self):
        with  open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system("git commit -m  \"first commit\"")
        os.system(f"git tag 0.1.0")
        os.system(f"git tag 0.2.0")
        with self.assertRaises(pygitversionhelper.gitversionhelper.tag.moreThanOneTag) :
            pygitversionhelper.gitversionhelper.tag.getLastTag(same_branch=True)
    """

    def test_defect__wrongargument_bump_type(self):
        _v = self._test_version_readback("0.1.1", version_std="PEP440")
        with self.assertRaises(pygitversionhelper.gitversionhelper.wrongArguments):
            pygitversionhelper.gitversionhelper.version.getCurrentVersion(bump_type="toto")

    def test_defect__wrongargument_bump_dev_strategy(self):
        _v = self._test_version_readback("0.1.1", version_std="PEP440")
        with self.assertRaises(pygitversionhelper.gitversionhelper.wrongArguments):
            pygitversionhelper.gitversionhelper.version.getCurrentVersion(bump_dev_strategy="toto")

    def test_nominal__tag__getDistanceFromTag(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.2.0")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue1")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue2")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.3.0")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue3")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue4")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        self.assertEqual(2, pygitversionhelper.gitversionhelper.tag.getDistanceFromTag())
        self.assertEqual(4, pygitversionhelper.gitversionhelper.tag.getDistanceFromTag("0.2.0"))

    def test_nominal__tag__getTags(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.2.0")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue1")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue2")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.3.0")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue3")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue4")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue5")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.3.1")

        self.assertEqual(
            set(["0.3.0", "0.2.0", "0.3.1"]),
            set(pygitversionhelper.gitversionhelper.tag.getTags()),
        )

    def test_nominal__tag__getTags_two_branch(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.2.0")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue1")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue2")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.3.0")

        os.system("git checkout -b dev")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue3")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue4")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue5")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.3.1")

        self.assertEqual(
            set(["0.3.0", "0.2.0", "0.3.1"]),
            set(pygitversionhelper.gitversionhelper.tag.getTags()),
        )

    def test_nominal__tag__getTags_two_branch_samebranch(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.2.0")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue1")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue2")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.3.0")

        os.system("git checkout -b dev")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue3")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue4")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue5")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.3.1")

        self.assertEqual(
            set(["0.3.0", "0.2.0", "0.3.1"]),
            set(pygitversionhelper.gitversionhelper.tag.getTags(same_branch=True)),
        )

        os.system("git switch master")

        self.assertEqual(
            set(["0.3.0", "0.2.0"]),
            set(pygitversionhelper.gitversionhelper.tag.getTags(same_branch=True)),
        )

        os.system("git switch dev")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue4")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.3.2")

        self.assertEqual(
            set(["0.3.0", "0.2.0", "0.3.1", "0.3.2"]),
            set(pygitversionhelper.gitversionhelper.tag.getTags(same_branch=True)),
        )
        self.assertEqual(
            set(["0.3.0", "0.2.0", "0.3.1", "0.3.2"]),
            set(pygitversionhelper.gitversionhelper.tag.getTags()),
        )

        os.system("git switch master")

        self.assertEqual(
            set(["0.3.0", "0.2.0"]),
            set(pygitversionhelper.gitversionhelper.tag.getTags(same_branch=True)),
        )
        self.assertEqual(
            set(["0.3.0", "0.2.0", "0.3.1", "0.3.2"]),
            set(pygitversionhelper.gitversionhelper.tag.getTags()),
        )

    def test_nominal__tag__getLastTag_two_branches(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system('git commit -m  "first commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue1")
        os.system("git add .")
        os.system('git commit -m  "2nd commit"')
        os.system(f"git tag 0.1.0")

        os.system("git checkout -b dev")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue2")
        os.system("git add .")
        os.system('git commit -m  "3rd commit"')
        os.system(f"git tag 0.2.0")

        """
        print("===")
        os.system(f"git status")
        print("===")
        os.system(f"git describe --tags --abbrev=0")
        print("===")
        os.system(f"git rev-list --all")
        print("===")
        os.system(f"git rev-list --tags")
        print("===")
        os.system(f"git log --oneline --decorate=short")
        print("===")
        os.system(f"git describe --tags")
        print("===")
        """

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(version_std="PEP440")
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.2.0", version_std="PEP440")
        self.assertEqual(_v.raw, "0.2.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 2)
        self.assertEqual(_v.patch, 0)

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue3")
        os.system("git add .")
        os.system('git commit -m  "3rd commit"')
        os.system(f"git tag 0.3.0")

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(version_std="PEP440")
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.3.0", version_std="PEP440")
        self.assertEqual(_v.raw, "0.3.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 3)
        self.assertEqual(_v.patch, 0)

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue4")
        os.system("git add .")
        os.system('git commit -m  "4th commit"')

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue5")
        os.system("git add .")
        os.system('git commit -m  "5th commit"')

        os.system("git switch master")

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(version_std="PEP440")
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.3.0", version_std="PEP440")
        self.assertEqual(_v.raw, "0.3.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 3)
        self.assertEqual(_v.patch, 0)

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(version_std="PEP440", same_branch=True)
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.1.0", version_std="PEP440")
        self.assertEqual(_v.raw, "0.1.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 0)

        os.system('git merge --no-ff dev -m "merge dev into master"')

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(version_std="PEP440", same_branch=True)
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.1.0", version_std="PEP440")
        self.assertEqual(_v.raw, "0.1.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 0)

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(version_std="PEP440")
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.3.0", version_std="PEP440")
        self.assertEqual(_v.raw, "0.3.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 3)
        self.assertEqual(_v.patch, 0)

        os.system(f"git tag 0.4.0")

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(version_std="PEP440", same_branch=True)
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.4.0", version_std="PEP440")
        self.assertEqual(_v.raw, "0.4.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 4)
        self.assertEqual(_v.patch, 0)

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(version_std="PEP440")
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.4.0", version_std="PEP440")
        self.assertEqual(_v.raw, "0.4.0")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 4)
        self.assertEqual(_v.patch, 0)

    def test_defect__tag__invalidtag(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue1")
        os.system("git add .")
        os.system('git commit -m  "2nd commit"')
        os.system(f"git tag INVALIDTAG")

        with self.assertRaises(pygitversionhelper.gitversionhelper.version.noValidVersion):
            _v = pygitversionhelper.gitversionhelper.version.getLastVersion()

        with self.assertRaises(pygitversionhelper.gitversionhelper.version.noValidVersion):
            _v = pygitversionhelper.gitversionhelper.version.getLastVersion(ignore_unknown_tags=True)

    def test_defect__tag__invalidtag_inbetween(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue1")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.0.1")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue2")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag INVALIDTAG")

        with self.assertRaises(pygitversionhelper.gitversionhelper.version.noValidVersion):
            _v = pygitversionhelper.gitversionhelper.version.getLastVersion()

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(ignore_unknown_tags=True)
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.0.1", version_std="PEP440")
        self.assertEqual(_v.raw, "0.0.1")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 0)
        self.assertEqual(_v.patch, 1)

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue3")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag INVALIDTAG2")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue4")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag INVALIDTA3")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue4")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag INVALIDTA4")

        with self.assertRaises(pygitversionhelper.gitversionhelper.version.noValidVersion):
            _v = pygitversionhelper.gitversionhelper.version.getLastVersion()

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(ignore_unknown_tags=True)
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.0.1", version_std="PEP440")
        self.assertEqual(_v.raw, "0.0.1")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 0)
        self.assertEqual(_v.patch, 1)

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue5")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.1.2")

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion()
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.1.2", version_std="PEP440")
        self.assertEqual(_v.raw, "0.1.2")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 2)

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(ignore_unknown_tags=True)
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.1.2", version_std="PEP440")
        self.assertEqual(_v.raw, "0.1.2")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 2)

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue6")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag INVALIDTA5")

        with self.assertRaises(pygitversionhelper.gitversionhelper.version.noValidVersion):
            _v = pygitversionhelper.gitversionhelper.version.getLastVersion()

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(ignore_unknown_tags=True)
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.1.2", version_std="PEP440")
        self.assertEqual(_v.raw, "0.1.2")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 2)

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue7")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        os.system(f"git tag 0.1.3")

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion()
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.1.3", version_std="PEP440")
        self.assertEqual(_v.raw, "0.1.3")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 3)

        _v = pygitversionhelper.gitversionhelper.version.getLastVersion(ignore_unknown_tags=True)
        self.assertIsInstance(_v, pygitversionhelper.gitversionhelper.version.MetaVersion)
        self._test_version_format(_v, "0.1.3", version_std="PEP440")
        self.assertEqual(_v.raw, "0.1.3")
        self.assertEqual(_v.major, 0)
        self.assertEqual(_v.minor, 1)
        self.assertEqual(_v.patch, 3)

    def test_nominal__commit_getLast(self):
        # sleeps are needed because commit date have a 1-sec accuracy :-/
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        time.sleep(1)

        initial_commit = pygitversionhelper.gitversionhelper.commit.getLast()

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue2")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        time.sleep(1)

        second_commit = pygitversionhelper.gitversionhelper.commit.getLast()

        self.assertNotEqual(initial_commit, second_commit)

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue3")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        time.sleep(1)

        third_commit = pygitversionhelper.gitversionhelper.commit.getLast()

        self.assertNotEqual(initial_commit, third_commit)
        self.assertNotEqual(second_commit, third_commit)

        os.system("git checkout -b dev")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue4")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        time.sleep(1)

        fourth_commit = pygitversionhelper.gitversionhelper.commit.getLast()

        self.assertNotEqual(initial_commit, fourth_commit)
        self.assertNotEqual(second_commit, fourth_commit)
        self.assertNotEqual(third_commit, fourth_commit)

        os.system("git switch master")

        fourth2_commit = pygitversionhelper.gitversionhelper.commit.getLast()
        self.assertEqual(fourth2_commit, fourth_commit)

        fourth3_commit = pygitversionhelper.gitversionhelper.commit.getLast(same_branch=True)

        self.assertNotEqual(fourth3_commit, fourth_commit)
        self.assertEqual(fourth3_commit, third_commit)

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue5")
        os.system("git add .")
        os.system('git commit -m  "commit"')
        time.sleep(1)

        fifth_commit = pygitversionhelper.gitversionhelper.commit.getLast()
        self.assertNotEqual(fifth_commit, fourth3_commit)

        os.system("git switch dev")

        fifth2_commit = pygitversionhelper.gitversionhelper.commit.getLast()
        self.assertEqual(fifth2_commit, fifth_commit)

        fifth2_commit = pygitversionhelper.gitversionhelper.commit.getLast(same_branch=True)
        self.assertEqual(fifth2_commit, fourth_commit)

    def test_defect__commit_notfound(self):

        with self.assertRaises(pygitversionhelper.gitversionhelper.commit.commitNotFound):
            pygitversionhelper.gitversionhelper.commit.getLast()

        with self.assertRaises(pygitversionhelper.gitversionhelper.commit.commitNotFound):
            pygitversionhelper.gitversionhelper.commit.getLast(same_branch=True)

    def test_defect__commit_getMessage_notfound(self):
        with self.assertRaises(pygitversionhelper.gitversionhelper.commit.commitNotFound):
            pygitversionhelper.gitversionhelper.commit.getMessage("")

    def test_nominal__commit_getFromTag(self):
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        initial_commit = pygitversionhelper.gitversionhelper.commit.getLast()

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue2")
        os.system("git add .")
        os.system('git commit -m  "commit2"')
        os.system(f"git tag TAG")

        second_commit = pygitversionhelper.gitversionhelper.commit.getLast()
        self.assertNotEqual(second_commit, initial_commit)
        second_commit2 = pygitversionhelper.gitversionhelper.commit.getFromTag("TAG")
        self.assertEqual(second_commit, second_commit2)
        self.assertNotEqual(second_commit2, initial_commit)

    def test_defect__commit_notfound2(self):

        with self.assertRaises(pygitversionhelper.gitversionhelper.commit.commitNotFound):
            pygitversionhelper.gitversionhelper.commit.getFromTag("TAG")

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system('git commit -m  "commit"')

        with self.assertRaises(pygitversionhelper.gitversionhelper.commit.commitNotFound):
            pygitversionhelper.gitversionhelper.commit.getFromTag("TAG")

        os.system(f"git tag OTHER_TAG")

        with self.assertRaises(pygitversionhelper.gitversionhelper.commit.commitNotFound):
            pygitversionhelper.gitversionhelper.commit.getFromTag("TAG")

    def test_nominal__commit_getMessage(self):
        commit_message = "AAAABBB CCCCDDDD".replace("\r\n", "\n").replace("\n", "\r\n")
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")
        os.system(f'git commit -m  "{commit_message}"')
        commit = pygitversionhelper.gitversionhelper.commit.getLast()
        message = pygitversionhelper.gitversionhelper.commit.getMessage(commit)
        self.assertEqual(message, commit_message)

    def test_nominal__commit_getMessage2(self):
        commit_message = """AAAABBB
        CCCCDDDD
        -f dfsds dfsdfs  $""".replace(
            "\r\n", "\n"
        ).replace(
            "\n", "\r\n"
        )

        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")

        cmd = "git commit -m".split()
        cmd.append(commit_message)
        subprocess.run(cmd, text=True, check=True)

        commit = pygitversionhelper.gitversionhelper.commit.getLast()
        message = pygitversionhelper.gitversionhelper.commit.getMessage(commit)

        self.assertEqual(message, commit_message)

    def test_nominal__commit_getMessagesSinceLastTag(self):
        commit_message1 = "1.1 update this" + os.linesep + "1.1 fix that" + os.linesep + "1.1 test"
        commit_message1 = commit_message1.replace("\r\n", "\n").replace("\n", "\r\n")
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue")
        os.system("git add .")

        cmd = "git commit -m".split()
        cmd.append(commit_message1)
        subprocess.run(cmd, text=True, check=True)
        os.system(f"git tag 0.1.1")

        commit_message2 = "2.1 update this" + os.linesep + "2.1 fix that" + os.linesep + "2.1 test"
        commit_message2 = commit_message2.replace("\r\n", "\n").replace("\n", "\r\n")
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue2")
        os.system("git add .")

        cmd = "git commit -m".split()
        cmd.append(commit_message2)
        subprocess.run(cmd, text=True, check=True)

        commit_message3 = "3.1 update this" + os.linesep + "3.1 fix that" + os.linesep + "3.1 test"
        commit_message3 = commit_message3.replace("\r\n", "\n").replace("\n", "\r\n")
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue3")
        os.system("git add .")

        cmd = "git commit -m".split()
        cmd.append(commit_message3)
        subprocess.run(cmd, text=True, check=True)

        commit_message4 = "4.1 update this" + os.linesep + "4.1 fix that" + os.linesep + "4.1 test"
        commit_message4 = commit_message4.replace("\r\n", "\n").replace("\n", "\r\n")
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue4")
        os.system("git add .")

        cmd = "git commit -m".split()
        cmd.append(commit_message4)
        subprocess.run(cmd, text=True, check=True)

        res = pygitversionhelper.gitversionhelper.commit.getMessagesSinceTag("0.1.1")
        self.assertEqual([commit_message4, commit_message3, commit_message2], res)

        res = pygitversionhelper.gitversionhelper.commit.getMessagesSinceTag("0.1.1", merged_output=True)
        self.assertEqual(os.linesep.join([commit_message4, commit_message3, commit_message2]), res)

        cmd = "git checkout -b dev".split()
        subprocess.run(cmd, text=True, check=True)

        commit_message5 = "5.1 update this" + os.linesep + "5.1 fix that" + os.linesep + "5.1 test"
        commit_message5 = commit_message5.replace("\r\n", "\n").replace("\n", "\r\n")
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue5")
        os.system("git add .")

        cmd = "git commit -m".split()
        cmd.append(commit_message5)
        subprocess.run(cmd, text=True, check=True)
        os.system(f"git tag 0.1.1.post4")

        commit_message6 = "6.1 update this" + os.linesep + "6.1 fix that" + os.linesep + "6.1 test"
        commit_message6 = commit_message6.replace("\r\n", "\n").replace("\n", "\r\n")
        with open("demofile.txt", "w+t") as tmpFile:
            tmpFile.write("testvalue6")
        os.system("git add .")

        cmd = "git commit -m".split()
        cmd.append(commit_message6)
        subprocess.run(cmd, text=True, check=True)

        cmd = "git switch master".split()
        subprocess.run(cmd, text=True, check=True)

        res = pygitversionhelper.gitversionhelper.commit.getMessagesSinceTag("0.1.1", merged_output=True, same_branch=True)
        self.assertEqual(os.linesep.join([commit_message4, commit_message3, commit_message2]), res)

        res = pygitversionhelper.gitversionhelper.commit.getMessagesSinceTag("0.1.1", merged_output=True)
        self.assertEqual(os.linesep.join([commit_message6, commit_message5, commit_message4, commit_message3, commit_message2]), res)

        time.sleep(1)

        merge_message = "automerge"
        cmd = "git merge --no-ff dev -m".split()
        cmd.append(merge_message)
        subprocess.run(cmd, text=True, check=True)

        res = pygitversionhelper.gitversionhelper.commit.getMessagesSinceTag("0.1.1", merged_output=True, same_branch=True)
        self.assertEqual(os.linesep.join([merge_message, commit_message4, commit_message3, commit_message2]), res)

        res = pygitversionhelper.gitversionhelper.commit.getMessagesSinceTag("0.1.1", merged_output=True)
        self.assertEqual(
            set(
                merge_message.splitlines()
                + commit_message6.splitlines()
                + commit_message5.splitlines()
                + commit_message4.splitlines()
                + commit_message3.splitlines()
                + commit_message2.splitlines()
            ),
            set(res.splitlines()),
        )

        res = pygitversionhelper.gitversionhelper.commit.getMessagesSinceTag("0.1.1", merged_output=True, ignore_merged=True)
        self.assertEqual(
            set(
                commit_message6.splitlines()
                + commit_message5.splitlines()
                + commit_message4.splitlines()
                + commit_message3.splitlines()
                + commit_message2.splitlines()
            ),
            set(res.splitlines()),
        )

    def tearDown(self):
        os.chdir("/")


class Test_gitversionhelperNoRepo(unittest.TestCase):
    def setUp(self):
        self.TmpWorkingDir = tempfile.TemporaryDirectory()
        self.TmpWorkingDirPath = pathlib.Path(self.TmpWorkingDir.name)
        os.chdir(self.TmpWorkingDirPath)

    def test_defect__norepo(self):
        with self.assertRaises(pygitversionhelper.gitversionhelper.repository.notAGitRepository):
            _v = pygitversionhelper.gitversionhelper.version.getCurrentVersion()

    def tearDown(self):
        os.chdir("/")
